from __future__ import unicode_literals

import base64
import json
import os
from multiprocessing import current_process

from django.conf import settings
from django.utils import timezone

from rest_framework.renderers import JSONRenderer

from eVotUM.Cripto import (pkiutils,
                           eccblind,
                           shamirsecret)
from . import routers
from .models import ElectionResults


def async_scrutinize(obj, data):
    """ Generate the election results """
    try:
        # Save state data
        p = current_process()
        obj.worker_pid = p.pid
        obj.result_status_choices = ElectionResults.ResultStatus.\
            SIGNATURE
        obj.save()
        # Reusable data
        election_id = obj.election_id
        results = {
            'candidates': {},
            'blank_votes': {},
            'total_votes': 0,
            'total_valid': 0,
            'invalid_signature': 0,
            'invalid_cipher': 0,
            'invalid_format': 0,
        }
        # Generate the password from the shared secret components
        password = fuse_shared_keys(
            data['secret_list'],
            data['elec_proc_id'],
            base64.b64decode(data['sge_cert'])
        )

        # Fetch the votes from the AS
        as_router = routers.ASrouter()
        more = True
        page = 1
        while more:
            batch_votes = []
            (votes, more) = as_router.fetch_votes(
                                election_id,
                                page)
            # Process each received vote
            for vote in votes:
                # Update the flat number of total votes
                results['total_votes'] += 1

                # Decypher the vote
                ciphered_vote = str(vote['ciphered_vote'])
                vs_signed_vote = vote['signed_vote_hash']
                # Check valid vote
                vs_pem = base64.b64decode(data['vs_cert'])
                (err, valid) = eccblind.verifySignature(
                    vs_pem,
                    vs_signed_vote,
                    vote['blind_components'],
                    vote['pr_components'],
                    ciphered_vote
                )
                # If vote is invalid
                # Send to the invalid votes list and continue
                if err is not None:
                    print err
                    results['invalid_signature'] += 1
                    continue

                # Decipher the vote
                (err, deciphered_vote) = pkiutils.decryptWithPrivateKey(
                    ciphered_vote,
                    base64.b64decode(data['election_pvt_key']),
                    password
                )
                # If the vote was not properly deciphered
                # Send to the invalid votes list and continue
                if err is not None:
                    print err
                    results['invalid_cipher'] += 1
                    continue

                try:
                    # Format sanitization
                    deciphered_vote = deciphered_vote.replace('\'', "\"")
                    # Check valid vote structure JSON
                    vote_dict = json.loads(deciphered_vote)

                    # Sign the vote and set the timestamp
                    (err, signed_vote) = pkiutils.signObject(
                        json.dumps(vote_dict),
                        base64.b64decode(data['cs_pvt_key']),
                        get_private_key_password()
                    )

                    if err is not None:
                        raise Exception("(%s) Could not sign vote" % err)

                    # Generate the vote dictionary
                    updated_vote = {}
                    updated_vote["id"] = vote.pop('id')
                    updated_vote["deciphered_vote"] = signed_vote
                    updated_vote["decipher_timestamp"] = str(timezone.now())

                    # Set the appropriate results statistics
                    candidates = vote_dict['candidates']
                    category = vote_dict['category']
                    blank_vote = vote_dict['blank']

                    # Update the correct list
                    if(blank_vote):
                        # First vote with this category
                        key = 'blank_votes'
                        if(category in results[key]):
                            results[key][category] += 1
                        else:
                            results[key][category] = 1
                    else:
                        key = 'candidates'
                        for candidate in candidates:
                            if(category in results[key]):
                                results[key][category] += 1
                            else:
                                results[key][category] = 1

                    # Update the flat number of valid votes
                    results['total_valid'] += 1

                # Catch the exception for mal-formatted JSON vote
                except ValueError as e:
                    print e
                    results['invalid_format'] += 1
                    continue
                # Other exceptions must bubble up
                except Exception as e:
                    raise e

                # Set the vote to be updated
                batch_votes.append(updated_vote)

            # Update this batch of votes
            if batch_votes:
                update_status = as_router.sign_votes(election_id, batch_votes)

                if update_status is None and batch_votes:
                    raise Exception("Unable to update votes")

                # Set the invalid votes from this batch
                results['invalid_format'] += update_status['invalid']

                for vote_dict in update_status['failed']:
                    # Failed votes have additional info
                    # For now just mark them as invalid by format
                    # TODO: What to do with such info
                    results['invalid_format'] += 1

            # Process the next batch
            page += 1

        # Update the process state to VERIFICATION
        obj.result_status_choices = ElectionResults.ResultStatus.\
            VERIFICATION
        obj.save()

        obj.election_results = JSONRenderer().render(results)
        # Update the process state to PUBLISHED
        obj.result_status_choices = ElectionResults.ResultStatus.\
            PUBLISHED

        # Unset the worker PID
        obj.worker_pid = None
        # Finish the process
        obj.save()

    except Exception as e:
        # TODO: Log
        print e
        # Update the process state to ERROR
        obj.result_status_choices = ElectionResults.ResultStatus.\
            ERROR
        # Unset the worker PID
        obj.worker_pid = None
        obj.save()


def get_private_key_password():
    return settings.SECRET_KEY


def is_worker_running(pid):
    """ Checks if a process is running """
    if pid is None:
        return False
    pid = int(pid)
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def fuse_shared_keys(component_list, uid, certificate):
    """ Combine all secret key components to
        generate the password """
    try:
        for idx, comp in enumerate(component_list):
            component_list[idx] = str(comp)

        (error_code, password) = shamirsecret.\
            recoverSecretFromComponents(
                component_list,
                str(uid),
                certificate
            )
        if error_code is None:
            return password

        raise Exception("(%s) Unable to recover shared secret"
                        % error_code)

    except Exception as e:
        print e
        raise e
