from __future__ import unicode_literals

import base64

from multiprocessing import Process
from rest_framework.response import Response
from rest_framework.generics import (RetrieveAPIView,
                                     DestroyAPIView,
                                     GenericAPIView)
from rest_framework.renderers import JSONRenderer
from rest_framework import status
from django.db import connection

from eVotUM.Cripto import pkiutils

from . import serializers
from .utils import (is_worker_running,
                    async_scrutinize)

from .models import ElectionResults


class ResultsStatusView(RetrieveAPIView):
    """
    Return the given election scrutiny status
    """
    permission_classes = []
    serializer_class = serializers.ResultsStatusSerializer
    lookup_field = 'election_id'

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def get_queryset(self):
        election_id = self.kwargs.get('election_id') or None
        queryset = ElectionResults.objects.all().filter(
            election_id=election_id)

        return queryset


class ResultsCleanView(DestroyAPIView):
    """
    Cleanup all results for the given election
    """
    permission_classes = []
    serializer_class = serializers.ResultsStatusSerializer
    lookup_field = 'election_id'

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def get_queryset(self):
        election_id = self.kwargs.get('election_id') or None
        queryset = ElectionResults.objects.all().filter(
            election_id=election_id)

        return queryset


class ResultsStartView(GenericAPIView):
    """
    Initialize all election scrutiny operations to
    generate the election results
    """
    permission_classes = []
    serializer_class = serializers.GenerateResultsSerializer
    lookup_field = 'election_id'

    def get_serializer_class(self):
        """
        Define the Serializer class to be interpreted
        by Django Swagger
        """
        return self.serializer_class

    def post(self, request, *args, **kwargs):
        return_data = {"error": False, "message": False}
        try:
            result_error = ElectionResults.ResultStatus.ERROR
            published_state = ElectionResults.ResultStatus.PUBLISHED
            # Serialize data
            serializer = serializers.GenerateResultsSerializer(
                data=request.data
            )
            serializer.is_valid(raise_exception=True)
            sanitized_data = serializer.validated_data

            # Get the election ID from the URL
            election_id = self.kwargs.get('election_id') or None

            # Check if an entry for this election already exists
            found = ElectionResults.objects.all().filter(
                election_id=election_id).count()

            # If already exists handle the error
            if found:
                result_obj = ElectionResults.objects.get(
                                election_id=election_id)

                # Avoiding deadlocks, check system consistency
                if (result_obj.result_status_choices is not result_error):
                    # If the state is not PUBLISHED, check for the worker PID
                    if result_obj.result_status_choices is published_state:
                        # Return the published results
                        return_data["message"] = result_obj.election_results
                        return Response(JSONRenderer().render(return_data))

                    # Otherwise check if the Worker is still alive
                    else:
                        if is_worker_running(result_obj.worker_pid):
                            raise Exception(
                                "Election results being generated")
                        else:
                            result_obj.result_status_choices = result_error
                            result_obj.save()
                            raise Exception(
                                "Last results generation process failed")

            else:
                # Initialize the ElectionResults entry
                result_obj = ElectionResults(election_id=election_id)
                result_obj.save()

            # Validate if Validator Service certificate is valid
            # and chained to the root certificate
            vs_cert = base64.b64decode(sanitized_data['vs_cert'])
            root_cert = base64.b64decode(sanitized_data['root_cert'])

            (err, vs_pub_key) = pkiutils.getPublicKeyFromCertificate(vs_cert)

            if err is not None:
                result_obj.result_status_choices = result_error
                result_obj.save()
                raise Exception("(%d) Invalid VS certificate" % err)

            # Validate if Root certificate is valid
            (err, root_pub_key) = pkiutils.\
                getPublicKeyFromCertificate(root_cert)

            if err is not None:
                result_obj.result_status_choices = result_error
                result_obj.save()
                raise Exception("(%d) Invalid Root certificate" % err)

            # Validate the certificate chain
            if not pkiutils.verifyCertificateChain(vs_cert, root_cert):
                result_obj.result_status_choices = result_error
                result_obj.save()
                raise Exception("VS certificate unchained")

            print vs_pub_key

            # Allow regenerating the DB connection on child processes
            connection.close()

            # Fork the process to allow asynchronous tasking
            p = Process(target=async_scrutinize,
                        args=(result_obj, sanitized_data,))
            p.daemon = True
            p.start()

            # Return an 'ok' message to the requestor
            return Response(JSONRenderer().render(return_data))

        except Exception as e:
            # TODO: Log
            return_data["error"] = True
            return_data["message"] = format(e)

            return Response(
                JSONRenderer().render(return_data),
                status=status.HTTP_400_BAD_REQUEST)
